import ai.h2o.featurestore.api.v1.CoreService_pb2 as pb

from ..entities.ingest import Ingest


class IngestHistory:
    def __init__(self, stub, feature_set):
        self._feature_set = feature_set
        self._stub = stub
        self._ingest_history = self.__load_history()

    def list(self):
        """Return a list of ingestion.

        Returns:
            list[Ingest]: A collection of ingestion.

        Typical example:
            history = my_feature_set.ingest_history()
            history.list()

        For more details:
            https://docs.h2o.ai/feature-store/latest-stable/docs/api/ingest_history_api.html#getting-the-ingestion-history
        """
        return [Ingest(self._stub, self._feature_set, ingest) for ingest in self._ingest_history]

    def refresh(self):
        """Refresh ingest history object with the latest ingestion (if there are any).

        Typical example:
            history = my_feature_set.ingest_history()
            history.refresh()
        """
        self._ingest_history = self.__load_history()

    @property
    def size(self):
        """A size of history."""
        return len(self._ingest_history)

    @property
    def first(self):
        """A first ingestion."""
        if self.list():
            return Ingest(self._stub, self._feature_set, self._ingest_history[0])
        else:
            raise Exception("No ingest has been performed so far.")

    @property
    def last(self):
        """A last ingestion."""
        if self.list():
            return Ingest(self._stub, self._feature_set, self._ingest_history[-1])
        else:
            raise Exception("No ingest has been performed so far.")

    def get(self, ingest_id):
        """Obtain a specific ingestion.

        Args:
            ingest_id: (str) A unique id of ingestion.

        Returns:
            Ingest: An existing ingestion.

        Typical example:
            history = my_feature_set.ingest_history()
            specific_ingest = history.get(ingest_id)

        Raises:
            Exception: No ingest has been found for the given ingest id.
        """
        ingests = [ingest for ingest in self._ingest_history if ingest.ingest_id == ingest_id]
        if ingests:
            return Ingest(self._stub, self._feature_set, ingests[0])
        else:
            raise Exception("No ingest has been found for the ingest id " + ingest_id)

    def __load_history(self):
        request = pb.GetIngestHistoryRequest()
        request.feature_set_id = self._feature_set.id
        request.feature_set_version = self._feature_set.version
        response = self._stub.GetIngestHistory(request)
        return response.ingest_history



import ai.h2o.featurestore.api.v1.CoreService_pb2 as pb

from ..utils import Utils
from .feature_set import FeatureSet
from .user import User


class FeatureSetReview:
    def __init__(self, stub, review):
        self._stub = stub
        self._review = review

    @property
    def review_id(self):
        return self._review.review_id

    @property
    def project_name(self):
        return self._review.project_name

    @property
    def feature_set_name(self):
        return self._review.feature_set_name

    @property
    def feature_set_major_version(self):
        return self._review.feature_set_major_version

    @property
    def feature_set_id(self):
        return self._review.feature_set_id

    @property
    def created_at(self):
        return Utils.timestamp_to_string(self._review.created_at)

    @property
    def status(self):
        return pb.ReviewStatus.Name(self._review.status)

    def __repr__(self):
        return Utils.pretty_print_proto(self._review)


class FeatureSetReviewRequest(FeatureSetReview):
    @property
    def author(self):
        return User(self._review.owner)

    def approve(self, reason):
        """Approve a review request.

        Args:
        reason: (str) A reason for review request approval.

        Typical example:
        review_request.approve("it will be fun")

        For more details:
        https://docs.h2o.ai/feature-store/latest-stable/docs/api/feature_set_review_api#manage-review-requests-from-other-users
        """
        request = pb.ApproveReviewRequest(review_id=self._review.review_id, reason=reason)
        self._stub.ApproveReview(request)

    def reject(self, reason):
        """Reject a review request.

        Args:
        reason: (str) A reason for review request rejection.

        Typical example:
        review_request.reject("it's not ready yet")

        For more details:
        https://docs.h2o.ai/feature-store/latest-stable/docs/api/feature_set_review_api#manage-review-requests-from-other-users
        """
        request = pb.RejectReviewRequest(review_id=self._review.review_id, reason=reason)
        self._stub.RejectReview(request)

    def get_feature_set(self) -> FeatureSet:
        """Get a feature set to review.

        Returns:
            A corresponding feature set.

        Typical example:
            review_request.get_feature_set()
        """
        request = pb.GetFeatureSetToReviewRequest(review_id=self._review.review_id)
        response = self._stub.GetFeatureSetToReview(request)
        return FeatureSet(self._stub, response.feature_set)

    def get_preview(self):
        """Preview the data of feature set to review.

        This previews up to a maximum of 100 rows and 50 features.

        Returns:
            list[dict]: A list of dictionary which contains JSON rows.

        Typical example:
            review_request.get_preview()
        """
        request = pb.GetFeatureSetPreviewToReviewRequest(
            review_id=self._review.review_id,
        )
        response = self._stub.GetFeatureSetPreviewToReview(request)
        if response.preview_url:
            json_response = Utils.fetch_preview_as_json_array(response.preview_url)
            return json_response
        else:
            return []


class FeatureSetUserReview(FeatureSetReview):
    @property
    def reviewer(self):
        return User(self._review.reviewer)

    @property
    def reviewed_at(self):
        return Utils.timestamp_to_string(self._review.reviewed_at)

    @property
    def reason(self):
        return self._review.reason

    def get_feature_set(self) -> FeatureSet:
        """Get a feature set in review.

        Returns:
            A corresponding feature set.

        Typical example:
            review_request.get_feature_set()

        For more details:
        https://docs.h2o.ai/feature-store/latest-stable/docs/api/feature_set_review_api#manage-own-feature-sets-in-review
        """
        request = pb.GetFeatureSetInReviewRequest(review_id=self._review.review_id)
        response = self._stub.GetFeatureSetInReview(request)
        return FeatureSet(self._stub, response.feature_set)

    def get_preview(self):
        """Preview the data of feature set in review.

        This previews up to a maximum of 100 rows and 50 features.

        Returns:
            list[dict]: A list of dictionary which contains JSON rows.

        Typical example:
            review_request.get_preview()

        For more details:
        https://docs.h2o.ai/feature-store/latest-stable/docs/api/feature_set_review_api#manage-own-feature-sets-in-review
        """
        request = pb.GetFeatureSetPreviewInReviewRequest(
            review_id=self._review.review_id,
        )
        response = self._stub.GetFeatureSetPreviewInReview(request)
        if response.preview_url:
            json_response = Utils.fetch_preview_as_json_array(response.preview_url)
            return json_response
        else:
            return []

    def delete(self):
        """Delete the currently major version in review.

        Review must be in status IN_PROGRESS or REJECTED.

        Typical example:
            review_request.delete()

        For more details:
        https://docs.h2o.ai/feature-store/latest-stable/docs/api/feature_set_review_api#manage-own-feature-sets-in-review
        """
        request = pb.DeleteRejectedFeatureSetRequest(review_id=self._review.review_id)
        self._stub.DeleteFeatureSetVersionInReview(request)



import re
from typing import List

from ai.h2o.featurestore.api.v1 import CoreService_pb2 as pb
from ai.h2o.featurestore.api.v1 import FeatureSetProtoApi_pb2 as FeatureSetApi

from .transformations import Transformation


class FeatureSchemaSpecialData:
    def __init__(self, spi=False, pci=False, rpi=False, demographic=False, sensitive=False):
        self.spi = spi
        self.pci = pci
        self.rpi = rpi
        self.demographic = demographic
        self.sensitive = sensitive

    def __repr__(self):
        return (
            f"spi={self.spi}, pci={self.pci}, rpi={self.rpi}, demographic={self.demographic}, "
            f"sensitive={self.sensitive}"
        )


class FeatureSchemaMonitoring:
    def __init__(self, anomaly_detection=False):
        self.anomaly_detection = anomaly_detection

    def __repr__(self):
        return f"anomaly_detection={self.anomaly_detection}"


class FeatureSchema:
    def __init__(
        self,
        name,
        _data_type,
        special_data=None,
        nested_features_schema=None,
        _feature_type="AUTOMATIC_DISCOVERY",
        description="",
        classifiers: set = None,
        custom_data=None,
        monitoring=None,
    ):
        self._data_type = _data_type
        self._name = name
        self._special_data = special_data or FeatureSchemaSpecialData()
        self._nested_features_schema = nested_features_schema or []
        self._schema = Schema(self._nested_features_schema, False)
        self._validated_feature_type = FeatureSchema.get_and_validate_feature_type(_feature_type)
        self._description = description
        self._classifiers = classifiers or set()
        self._custom_data = custom_data
        self._monitoring = monitoring or FeatureSchemaMonitoring()

    @property
    def schema(self):
        return self._schema

    @property
    def special_data(self):
        return self._special_data

    @special_data.setter
    def special_data(self, value):
        self._special_data = value

    @property
    def name(self):
        return self._name

    @property
    def data_type(self):
        if self._schema.size() > 0:
            nested_data_type = ", ".join(
                ["{}: {}".format(nested.name, nested.data_type) for nested in self._nested_features_schema]
            )
            return self._data_type.format(nested_data_type)
        else:
            return self._data_type

    @data_type.setter
    def data_type(self, value):
        self._data_type = value

    @property
    def feature_type(self):
        return FeatureSetApi.FeatureType.Name(self._validated_feature_type)

    @feature_type.setter
    def feature_type(self, value):
        self._validated_feature_type = FeatureSchema.get_and_validate_feature_type(value)

    @property
    def description(self):
        return self._description

    @description.setter
    def description(self, value):
        self._description = value

    @property
    def classifiers(self):
        return self._classifiers

    @classifiers.setter
    def classifiers(self, value: set):
        self._classifiers = value

    @property
    def custom_data(self):
        return self._custom_data

    @custom_data.setter
    def custom_data(self, value):
        self._custom_data = value

    @property
    def monitoring(self):
        return self._monitoring

    @monitoring.setter
    def monitoring(self, value):
        self._monitoring = value

    def to_string(self):
        return self._to_string("")

    def _to_string(self, prefix):
        initial = "- " if prefix == "" else prefix
        res = f"{initial}{self.name}: {self.data_type}"
        if self._schema.size() > 0:
            nested_prefix = "  |-- " if prefix == "" else "  " + prefix
            res += "\n" + self._schema._to_string(nested_prefix)
        return res

    @staticmethod
    def create_from(feature):
        nested = [FeatureSchema.create_from(nested_feature) for _, nested_feature in feature.nested_features.items()]
        return FeatureSchema(
            feature.name,
            feature.data_type,
            FeatureSchemaSpecialData(
                feature.special_data.spi,
                feature.special_data.pci,
                feature.special_data.rpi,
                feature.special_data.demographic,
                feature.special_data.sensitive,
            ),
            nested,
            feature.profile.feature_type,
            feature.description,
            feature.classifiers,
            feature.custom_data,
            FeatureSchemaMonitoring(feature.monitoring.anomaly_detection),
        )

    @staticmethod
    def get_and_validate_feature_type(feature_type):
        valid_values = FeatureSetApi.FeatureType.keys()
        if feature_type.upper() in valid_values:
            FeatureSetApi.FeatureType.Value(feature_type.upper())
        else:
            raise Exception("Invalid feature type. Supported values are: " + ", ".join(map(str, valid_values)))

    def __repr__(self):
        return self.to_string()


class VersionedId:
    def __init__(self, id: str, major_version: int):
        self.id = id
        self.major_version = major_version


class SchemaDerivation:
    def __init__(
        self,
        feature_set_ids: List[VersionedId],
        transformation: Transformation,
    ):
        self.feature_set_ids = feature_set_ids
        self.transformation = transformation

    def _to_proto(self):
        return pb.DerivedInformation(
            feature_set_ids=[pb.VersionedId(id=f.id, major_version=f.major_version) for f in self.feature_set_ids],
            transformation=self.transformation._to_proto(),
        )


class Schema:
    def __init__(self, feature_schemas, is_root_schema, derivation: SchemaDerivation = None):
        self._feature_schemas = feature_schemas
        self._is_root_schema = is_root_schema
        if is_root_schema:
            self._field_separator = " "
        else:
            self._field_separator = ": "
        self.derivation = derivation

    def __getitem__(self, feature_name):
        filtered = list(filter(lambda f: f.name == feature_name, self._feature_schemas))
        if len(filtered):
            return filtered[0]
        else:
            raise Exception(f"Feature name '{feature_name}' does not exist in the schema.")

    @staticmethod
    def create_from(obj):
        """Create a schema from String and FeatureSet.

        Args:
            obj: (str | FeatureSet) Object represents String or FeatureSet.

        Returns:
            Schema: A schema with feature names and data types.

            For example:

            id INT, text STRING, label DOUBLE, state STRING, date TIMESTAMP

        Typical example:
            schema = Schema.create_from("col1 string, col2 string, col3 integer")

        Raises:
            Exception: Schema can be created either from string or existing feature set.

        For more details:
            https://docs.h2o.ai/feature-store/latest-stable/docs/api/schema_api.html#create-a-schema-from-a-string
        """
        from .entities.feature_set import FeatureSet

        if isinstance(obj, FeatureSet):
            features = [FeatureSchema.create_from(feature) for name, feature in obj.features.items()]
            if obj.is_derived():
                derived_from = obj._feature_set.derived_from
                derivation = SchemaDerivation(
                    [VersionedId(f.id, f.major_version) for f in derived_from.feature_set_ids],
                    Transformation.from_proto(derived_from.transformation),
                )
                return Schema(features, True, derivation)
            else:
                return Schema(features, True)
        elif isinstance(obj, str):
            return Schema(Schema._parse_features_schema(obj, True)[::-1], True)
        else:
            raise Exception("Schema can be created either from string or existing feature set.")

    @staticmethod
    def create_derived_from(schema_string: str, derive_from, transformation: Transformation):
        """Create a schema from a string formatted schema and parent feature set along with transformation.

        Args:
            schema_string: (str) String contains column names and data types.
            derive_from: (FeatureSet) Represents a feature set.
            transformation: (Transformation) Represents an instance of Transformation.

        Returns:
            Schema: A schema with feature names and data types.

            For example:

            id INT, text STRING, label DOUBLE, state STRING, date TIMESTAMP
        """
        return Schema(
            Schema._parse_features_schema(schema_string, True)[::-1],
            True,
            SchemaDerivation(
                [VersionedId(f.id, f.major_version) for f in derive_from],
                transformation,
            ),
        )

    def select(self, feature_names):
        """Create a new schema by column selection.

        Args:
            feature_names: (Iterable[str]) String or iterable collection of strings.

        Returns:
            Schema: A schema with selected feature names and data types.

            For example:

            id INT, text STRING, label DOUBLE, state STRING, date TIMESTAMP

        Typical example:
            schema_str = "id INT, text STRING, label DOUBLE, state STRING, date STRING"
            schema = Schema.create_from(schema_str)
            schema.select("id text label")

        For more details:
            https://docs.h2o.ai/feature-store/latest-stable/docs/api/schema_api.html#create-a-new-schema-by-column-selection
        """
        self._feature_schemas = list(filter(lambda schema: schema.name in feature_names, self._feature_schemas))
        return self

    def exclude(self, feature_names):
        """Create a new schema by column selection.

        Args:
            feature_names: (Iterable[str]) String or iterable collection of strings.

        Returns:
            Schema: A schema with feature names and data types without excluded names.

            For example:

            id INT, text STRING, label DOUBLE, state STRING, date TIMESTAMP

        Typical example:
            schema_str = "id INT, text STRING, label DOUBLE, state STRING, date STRING"
            schema = Schema.create_from(schema_str)
            schema.exclude("id text label")

        For more details:
            https://docs.h2o.ai/feature-store/latest-stable/docs/api/schema_api.html#create-a-new-schema-by-column-selection
        """
        self._feature_schemas = list(filter(lambda schema: schema.name not in feature_names, self._feature_schemas))
        return self

    def append(self, feature_schema, after=None):
        """Create a new schema.

        A schema is created by adding a new feature schema to the end / after the selected column.

        Args:
            feature_schema: (FeatureSchema) Represents new feature schema.
            after: (FeatureSchema) Represents an existing feature schema.

        Returns:
            Schema: A schema with existing feature names and data types and newly added feature column.

            For example:

            id INT, text STRING, label DOUBLE, state STRING, date TIMESTAMP

        Typical example:
            new_feature_schema = FeatureSchema("new_name", STRING)
            schema.append(new_feature_schema)  # Append to the end
            schema.append(new_feature_schema, schema["old"])  # Append after old

        Raises:
            Exception: Feature name does not exist in the schema.

        For more details:
            https://docs.h2o.ai/feature-store/latest-stable/docs/api/schema_api.html#create-a-new-schema-by-adding-a-new-feature-schema
        """
        if not after:
            self._feature_schemas.append(feature_schema)
        else:
            index = self._feature_schemas.index(after) + 1
            self._feature_schemas.insert(index, feature_schema)
        return self

    def prepend(self, feature_schema, before=None):
        """Create a new schema.

        A schema is created by adding a new feature schema to the beginning / before the selected column.

        Args:
            feature_schema: (FeatureSchema) Represents new feature schema.
            before: (FeatureSchema) Represents an existing feature schema.

        Returns:
            Schema: A schema with existing feature names and data types and newly added feature column.

            For example:

            id INT, text STRING, label DOUBLE, state STRING, date TIMESTAMP

        Typical example:
            new_feature_schema = FeatureSchema("new_name", STRING)
            schema.prepend(new_feature_schema)  # Prepend to the beginning
            schema.prepend(new_feature_schema, schema["old"])  # Prepend before old

        Raises:
            Exception: Feature name does not exist in the schema.

        For more details:
            https://docs.h2o.ai/feature-store/latest-stable/docs/api/schema_api.html#create-a-new-schema-by-adding-a-new-feature-schema
        """
        if not before:
            self._feature_schemas.insert(0, feature_schema)
        else:
            index = self._feature_schemas.index(before)
            self._feature_schemas.insert(index, feature_schema)
        return self

    def size(self):
        """Return the length of a schema."""
        return len(self._feature_schemas)

    def is_derived(self):
        return self.derivation is not None

    def to_string(self):
        return self._to_string("")

    def _to_string(self, prefix):
        return "\n".join(["{}".format(nested._to_string(prefix)) for nested in self._feature_schemas])

    def _to_proto_schema(self):
        return [
            pb.FeatureSchema(
                name=feature_schema.name,
                data_type=feature_schema.data_type,
                nested=feature_schema.schema._to_proto_schema(),
                special_data=pb.FeatureSchema.SpecialData(
                    spi=feature_schema.special_data.spi,
                    pci=feature_schema.special_data.pci,
                    rpi=feature_schema.special_data.rpi,
                    demographic=feature_schema.special_data.demographic,
                    sensitive=feature_schema.special_data.sensitive,
                ),
                feature_type=feature_schema._validated_feature_type,
                description=feature_schema.description,
                classifiers=list(feature_schema.classifiers),
                custom_data=feature_schema.custom_data,
                monitoring=pb.FeatureSchema.Monitoring(anomaly_detection=feature_schema.monitoring.anomaly_detection),
            )
            for feature_schema in self._feature_schemas
        ]

    def __repr__(self):
        return self.to_string()

    @staticmethod
    def _parse_features_schema(schema_string, is_root_level):
        if schema_string:
            feature_schema, remaining_schema_string = Schema._extract_pair(schema_string, is_root_level)
            return Schema._parse_features_schema(remaining_schema_string, is_root_level) + [feature_schema]
        else:
            return []

    @staticmethod
    def _get_index_of_matching_ending_angular_bracket(string):
        stack = []
        for idx, ch in enumerate(string):
            if ch == "<":
                stack.append(idx)
            elif ch == ">":
                if len(stack) == 0:
                    raise Exception("Invalid schema!")
                if len(stack) == 1:
                    return idx
                else:
                    stack.pop()
        if len(stack) != 0:
            raise Exception("Invalid schema!")

    @staticmethod
    def _extract_pair(schema, is_root_level):
        if is_root_level:
            splits = schema.split(maxsplit=1)
            remaining_with_data_type = splits[1]
        else:
            splits = re.split(r"\s*:\s*", schema, maxsplit=1)
            remaining_with_data_type = splits[1]
        col_name = splits[0]
        if Schema._is_complex(remaining_with_data_type):
            idx = Schema._get_index_of_matching_ending_angular_bracket(remaining_with_data_type) + 1
            if idx is None:
                raise Exception("Invalid schema.")
            nested_data_type, remaining_schema = (
                remaining_with_data_type[:idx],
                remaining_with_data_type[idx:],
            )
            if remaining_with_data_type.upper().startswith("STRUCT"):
                nested_schema, data_type = Schema._extract_struct(nested_data_type)
            else:
                nested_schema, data_type = Schema._extract_array(nested_data_type)
            rest = remaining_schema
        else:
            data_type, rest = Schema._extract_primitive(remaining_with_data_type)
            nested_schema = []
        parsed_feature_schema = FeatureSchema(col_name, data_type, nested_features_schema=nested_schema)
        remaining_unparsed = re.sub(r"^\s*,?\s*", "", rest)
        return parsed_feature_schema, remaining_unparsed

    @staticmethod
    def _is_complex(data_type):
        first_non_array_type = Schema._extract_first_non_array_data_type(data_type)
        return first_non_array_type.upper().startswith("STRUCT")

    @staticmethod
    def _extract_first_non_array_data_type(data_type):
        if data_type.upper().startswith("ARRAY"):
            nested_data_type = data_type.strip()[len("ARRAY") : len(data_type)].strip()[1:-1]
            return Schema._extract_first_non_array_data_type(nested_data_type)
        else:
            return data_type

    @staticmethod
    def _extract_struct(data_type):
        data_type_content = data_type.strip()[len("STRUCT") : len(data_type)].strip()[1:-1]
        return (
            Schema._parse_features_schema(data_type_content, is_root_level=False)[::-1],
            "STRUCT<{0}>",
        )

    @staticmethod
    def _extract_array(data_type):
        data_type_content = data_type.strip()[len("ARRAY") : len(data_type)].strip()[1:-1]
        if data_type_content.upper().startswith("ARRAY"):
            nested_schema, placeholder = Schema._extract_array(data_type_content)
            return nested_schema, f"ARRAY<{placeholder}>"
        elif data_type_content.upper().startswith("STRUCT"):
            nested_schema, placeholder = Schema._extract_struct(data_type_content)
            return nested_schema, f"ARRAY<{placeholder}>"
        else:
            return [], ""

    @staticmethod
    def _extract_primitive(schema):
        splits = re.split(r"\s*,\s*", schema, maxsplit=1)
        data_type = splits[0].strip()
        # Call bellow is done for validation purposes
        Schema._get_index_of_matching_ending_angular_bracket(data_type)
        if len(splits) > 1:
            remaining_schema = splits[1].strip()
        else:
            remaining_schema = ""
        return data_type, remaining_schema



from ai.h2o.featurestore.api.v1 import CoreService_pb2 as pb

IN_PROGRESS = pb.ReviewStatus.REVIEW_STATUS_TO_REVIEW
APPROVED = pb.ReviewStatus.REVIEW_STATUS_APPROVED
REJECTED = pb.ReviewStatus.REVIEW_STATUS_REJECTED



import ai.h2o.featurestore.api.v1.CoreService_pb2 as pb
import ai.h2o.featurestore.api.v1.FeatureSetProtoApi_pb2 as FeatureSetApi

from ..utils import Utils


class MLDatasetFeature:
    def __init__(self, internal_feature: pb.MLDatasetFeature):
        self._internal_feature = internal_feature

    @property
    def name(self):
        return self._internal_feature.name

    @property
    def feature_type(self):
        return FeatureSetApi.FeatureType.Name(self._internal_feature.feature_type)

    @property
    def data_type(self):
        return self._internal_feature.data_type

    @property
    def is_primary_key(self):
        return self._internal_feature.is_primary_key

    def __repr__(self):
        return Utils.pretty_print_proto(self._internal_feature)

    def __str__(self):
        return (
            f"Name                      : {self.name} \n"
            f"Original Name             : {self._internal_feature.original_name} \n"
            f"Original data type        : {self._internal_feature.original_data_type} \n"
            f"Is primary key?           : {self.is_primary_key} \n"
            f"Transformation function   : {self._internal_feature.transformation_function} \n"
            f"Data type                 : {self.data_type} \n"
            f"Feature type              : {self.feature_type} \n"
            f"Feature view version      : {self._internal_feature.feature_view_version} \n"
            f"Feature set major version : {self._internal_feature.feature_set_major_version} \n"
        )



from __future__ import annotations

from typing import TYPE_CHECKING, Union

import ai.h2o.featurestore.api.v1.CoreService_pb2 as pb

from ..utils import Utils
from .feature_ref import FeatureRef

if TYPE_CHECKING:  # Only imports the below statements during type checking
    from .feature import Feature


class TransformationFunctionMetadata:
    def __init__(self, transformation_function_proto: pb.TransformationFunction):
        self._transformation_function = transformation_function_proto

    @property
    def function_name(self):
        return self._transformation_function.function_name

    @property
    def supported_input_types(self):
        return self._transformation_function.supported_input_types

    @property
    def output_type(self):
        return self._transformation_function.output_type

    def apply(self, feature: Union[Feature, FeatureRef]):
        return TransformationFunction(self.function_name).apply(feature)

    def __repr__(self):
        return Utils.pretty_print_proto(self._transformation_function)

    def __str__(self):
        return (
            f"Function name         : {self.function_name} \n"
            f"Supported input types : {self.supported_input_types} \n"
            f"Output type           : {self.output_type} \n"
        )


class TransformationFunction:
    def __init__(self, function_name: str):
        self._function_name = function_name

    @property
    def function_name(self):
        return self._function_name

    def apply(self, feature: Union[Feature, FeatureRef]):
        return TransformedFeature(self, feature)

    def __repr__(self):
        return self._function_name


class TransformedFeature:
    def __init__(
        self,
        transformation: TransformationFunction,
        feature: Union[Feature, FeatureRef],
    ):
        self.transformation = transformation
        self.feature = feature if isinstance(feature, FeatureRef) else feature._reference()



import os
from os.path import expanduser

from jproperties import Properties


class ConfigUtils:
    TOKEN_KEY = "token"
    INTERACTIVE_LOGGING = "progress.logging.console"
    _props = Properties()

    @staticmethod
    def collect_properties():
        file = ConfigUtils.get_properties_path()
        if not os.path.isfile(file):
            open(file, "a").close()
        with open(file, "rb") as f:
            ConfigUtils._props.load(f, "utf-8")
        return ConfigUtils._props

    @staticmethod
    def store_properties(props):
        with open(ConfigUtils.get_properties_path(), "wb") as f:
            props.store(f, encoding="utf-8")

    @staticmethod
    def delete_property(props, key):
        del props[key]
        ConfigUtils.store_properties(props)

    @staticmethod
    def store_property(props, key, value):
        props[key] = value
        ConfigUtils.store_properties(props)

    @staticmethod
    def set_property(props, key, value):
        props[key] = value
        ConfigUtils.store_properties(props)

    @staticmethod
    def get_property(key):
        return ConfigUtils._props[key].data

    @staticmethod
    def store_token(props, token):
        ConfigUtils.store_property(props, ConfigUtils.TOKEN_KEY, token)

    @staticmethod
    def get_token(props):
        return props[ConfigUtils.TOKEN_KEY].data

    @staticmethod
    def is_interactive_print_enabled():
        return (
            ConfigUtils._props.get(ConfigUtils.INTERACTIVE_LOGGING) is None
            or ConfigUtils._props[ConfigUtils.INTERACTIVE_LOGGING].data.lower() == "true"
        )

    @staticmethod
    def get_properties_path():
        if os.environ.get("FEATURESTORE_USER_CONFIG") is not None:
            return os.environ.get("FEATURESTORE_USER_CONFIG")
        else:
            return expanduser("~") + os.path.sep + ".featurestore.config"



import os
import shutil
import tempfile
from abc import ABC, abstractmethod
from enum import Enum

import requests

import ai.h2o.featurestore.api.v1.CoreService_pb2 as pb
from ai.h2o.featurestore.api.v1.CoreService_pb2_grpc import CoreServiceStub

from .utils import Utils


class Transformation(ABC):
    @abstractmethod
    def _initialize(self, grpc: CoreServiceStub):
        raise NotImplementedError("Method `_initialize` needs to be implemented by the child class")

    @abstractmethod
    def _to_proto(self):
        raise NotImplementedError("Method `to_proto` needs to be implemented by the child class")

    @staticmethod
    def from_proto(proto: pb.Transformation):
        if proto.HasField("mojo"):
            mojo = DriverlessAIMOJO(None)
            mojo.mojo_remote_location = proto.mojo.filename
            return mojo
        elif proto.HasField("spark_pipeline"):
            spark_pipeline = SparkPipeline(None)
            spark_pipeline.pipeline_remote_location = proto.spark_pipeline.filename
            return spark_pipeline
        elif proto.HasField("join"):
            return JoinFeatureSets(
                proto.join.left_key, proto.join.right_key, JoinFeatureSetsType._from_proto(proto.join.join_type)
            )


class DriverlessAIMOJO(Transformation):
    def __init__(self, mojo_local_location):
        self.mojo_local_location = mojo_local_location
        self.mojo_remote_location = None

    def _initialize(self, grpc: CoreServiceStub):
        if self.mojo_remote_location is None:
            if not os.path.exists(self.mojo_local_location):
                raise Exception(f"Provided file ${self.mojo_local_location} doesn't exists.")

            md5_checksum = Utils.generate_md5_checksum(self.mojo_local_location)
            upload_response = grpc.GenerateTransformationUpload(
                pb.GenerateTransformationUploadRequest(
                    transformation_type=pb.TransformationType.TransformationMojo,
                    md5_checksum=md5_checksum,
                )
            )
            with open(self.mojo_local_location, "rb") as mojo_file:
                response = requests.put(
                    url=upload_response.url,
                    data=mojo_file,
                    headers=upload_response.headers,
                )
                if response.status_code not in range(200, 300):
                    raise Exception(
                        f"DriverlessAIMOJO file upload failed with status code {response.status_code} "
                        f"and message {response.text}"
                    )
                self.mojo_remote_location = upload_response.filename

    def _to_proto(self):
        return pb.Transformation(mojo=pb.MojoTransformation(filename=self.mojo_remote_location))


class SparkPipeline(Transformation):
    def __init__(self, pipeline):
        if pipeline:
            if isinstance(pipeline, str) and pipeline.endswith(".zip"):
                if not os.path.exists(pipeline):
                    raise Exception(f"Provided file {pipeline} doesn't exists.")
                self.pipeline_local_location = pipeline
            elif isinstance(pipeline, str):
                shutil.make_archive("pipeline", "zip", pipeline)
                self.pipeline_local_location = "pipeline.zip"
            else:
                if Utils.is_running_on_databricks():
                    import random
                    import string

                    output_dir = "/tmp/" + "".join((random.choice(string.ascii_lowercase) for x in range(10)))
                    remote_output_dir = "/dbfs" + output_dir
                else:
                    output_dir = tempfile.mkdtemp()
                    remote_output_dir = output_dir
                pipeline.write().overwrite().save(output_dir)
                shutil.make_archive("pipeline", "zip", remote_output_dir)
                shutil.rmtree(remote_output_dir)
                self.pipeline_local_location = "pipeline.zip"
        else:
            self.pipeline_local_location = pipeline
        self.pipeline_remote_location = None

    def _initialize(self, grpc: CoreServiceStub):
        if self.pipeline_remote_location is None:
            md5_checksum = Utils.generate_md5_checksum(self.pipeline_local_location)
            upload_response = grpc.GenerateTransformationUpload(
                pb.GenerateTransformationUploadRequest(
                    transformation_type=pb.TransformationType.TransformationSparkPipeline,
                    md5_checksum=md5_checksum,
                )
            )

            with open(self.pipeline_local_location, "rb") as spark_pipeline_file:
                response = requests.put(
                    url=upload_response.url,
                    data=spark_pipeline_file,
                    headers=upload_response.headers,
                )
                if response.status_code not in range(200, 300):
                    raise Exception(
                        f"SparkPipeline file upload failed with status code {response.status_code} "
                        f"and message {response.text}"
                    )
                self.pipeline_remote_location = upload_response.filename

    def _to_proto(self):
        return pb.Transformation(spark_pipeline=pb.SparkPipelineTransformation(filename=self.pipeline_remote_location))


class JoinFeatureSetsType(Enum):
    INNER = 0
    LEFT = 1
    RIGHT = 2
    FULL = 3
    CROSS = 4

    def _to_proto(self):
        return pb.JoinType.Value("JOIN_TYPE_" + self.name)

    @classmethod
    def _from_proto(cls, join_type: pb.JoinType):
        if join_type == pb.JOIN_TYPE_INNER:
            return JoinFeatureSetsType.INNER
        elif join_type == pb.JOIN_TYPE_LEFT:
            return JoinFeatureSetsType.LEFT
        elif join_type == pb.JOIN_TYPE_RIGHT:
            return JoinFeatureSetsType.RIGHT
        elif join_type == pb.JOIN_TYPE_FULL:
            return JoinFeatureSetsType.FULL
        elif join_type == pb.JOIN_TYPE_CROSS:
            return JoinFeatureSetsType.CROSS
        else:
            raise Exception("Unsupported join_type" + str(join_type))


class JoinFeatureSets(Transformation):
    def __init__(self, left_key, right_key, join_type: JoinFeatureSetsType = JoinFeatureSetsType.INNER):
        self.left_key = left_key
        self.right_key = right_key
        self.join_type = join_type

    def _initialize(self, grpc: CoreServiceStub):
        pass

    def _to_proto(self):
        return pb.Transformation(
            join=pb.JoinTransformation(
                left_key=self.left_key,
                right_key=self.right_key,
                join_type=self.join_type._to_proto(),
            )
        )



import datetime
from typing import List

from dateutil.tz import gettz
from google.protobuf.timestamp_pb2 import Timestamp

import ai.h2o.featurestore.api.v1.CoreService_pb2 as pb

from ..entities.pat import PersonalAccessToken


class PersonalAccessTokens:
    def __init__(self, stub):
        self._stub = stub

    def generate(self, name: str, description: str, expiry_date: str = None, timezone: str = None) -> str:
        """Generate a personal access token for the currently logged-in user.

        Args:
            name: (str) A token name.
            description: (str) A description about the token.
            expiry_date: (str) Object represents a date string with format dd/MM/yyyy. Default is None.
            timezone: (str) Object represents a time zone name (Eg: 'America/Chicago'). Default is None.

        Returns:
            str: A token string for authentication.

        Typical example:
            token_str = client.auth.pats.generate(name="background_jobs", description="some description",
              expiry_date="<dd/MM/yyyy>", timezone=None)

        Raises:
            Exception: Invalid timezone.
            ValueError: Expiry date must be in the format: dd/MM/yyyy.

        For more details:
            https://docs.h2o.ai/feature-store/latest-stable/docs/api/authentication.html#authentication-via-personal-access-tokens-pats
        """
        request = pb.GenerateTokenRequest()
        request.name = name
        request.description = description
        if expiry_date:
            try:
                if timezone:
                    desired_timezone = gettz(timezone)
                    if not desired_timezone:
                        raise Exception("Invalid timezone id: '{}'".format(timezone))
                else:
                    desired_timezone = None
                expiration = datetime.datetime.strptime(expiry_date, "%d/%m/%Y").astimezone(desired_timezone)
                timestamp = Timestamp()
                timestamp.FromDatetime(expiration)
                request.expiry_date.CopyFrom(timestamp)
            except ValueError:
                raise Exception("Expiry date must be in the format: dd/MM/yyyy")
        response = self._stub.GenerateToken(request)
        return response.token

    def list(self) -> List[PersonalAccessToken]:
        """Return a generator which obtains the personal access tokens owned by current user, lazily.

        Returns:
            Iterable[PersonalAccessToken]: A generator iterator object consists of personal access tokens.

        Typical example:
            client.auth.pats.list()
        """
        request = pb.ListPersonalAccessTokensRequest()
        while request:
            response = self._stub.ListPersonalAccessTokens(request)
            if response.next_page_token:
                request.page_token = response.next_page_token
            else:
                request = None
            for pat in response.personal_access_tokens:
                yield PersonalAccessToken(self._stub, pat)

    def get(self, token_id: str) -> PersonalAccessToken:
        """Obtain a particular personal access token.

        Args:
            token_id: (str) A unique id of a token object.

        Returns:
            PersonalAccessToken: A token object.

        Typical example:
            client.auth.pats.get("token_id")
        """
        request = pb.TokenRequest()
        request.token_id = token_id
        response = self._stub.GetToken(request)
        return PersonalAccessToken(self._stub, response.token)

    def __repr__(self):
        return "This class wraps together methods working with Personal Access Tokens (PATs)"



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: ai/h2o/featurestore/api/v1/FeatureSetSearch.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n1ai/h2o/featurestore/api/v1/FeatureSetSearch.proto\x12\x1a\x61i.h2o.featurestore.api.v1\"?\n\rTextualFilter\x12\r\n\x05\x66ield\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x03(\t\x12\x10\n\x08operator\x18\x03 \x01(\t\"A\n\x0fNumericalFilter\x12\r\n\x05\x66ield\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x01\x12\x10\n\x08operator\x18\x03 \x01(\t\"?\n\rBooleanFilter\x12\r\n\x05\x66ield\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x08\x12\x10\n\x08operator\x18\x03 \x01(\t\"\xce\x01\n\x06\x46ilter\x12\x39\n\x04text\x18\x01 \x01(\x0b\x32).ai.h2o.featurestore.api.v1.TextualFilterH\x00\x12>\n\x07numeric\x18\x02 \x01(\x0b\x32+.ai.h2o.featurestore.api.v1.NumericalFilterH\x00\x12<\n\x07\x62oolean\x18\x03 \x01(\x0b\x32).ai.h2o.featurestore.api.v1.BooleanFilterH\x00\x42\x0b\n\tconditionB\x17\x42\x15\x46\x65\x61tureSetSearchProtob\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'ai.h2o.featurestore.api.v1.FeatureSetSearch_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'B\025FeatureSetSearchProto'
  _globals['_TEXTUALFILTER']._serialized_start=81
  _globals['_TEXTUALFILTER']._serialized_end=144
  _globals['_NUMERICALFILTER']._serialized_start=146
  _globals['_NUMERICALFILTER']._serialized_end=211
  _globals['_BOOLEANFILTER']._serialized_start=213
  _globals['_BOOLEANFILTER']._serialized_end=276
  _globals['_FILTER']._serialized_start=279
  _globals['_FILTER']._serialized_end=485
# @@protoc_insertion_point(module_scope)



import configparser
import logging
import os
from urllib.parse import urlparse

import ai.h2o.featurestore.api.v1.CoreService_pb2 as pb

from .user_credentials import (
    AzureKeyCredentials,
    AzurePrincipalCredentials,
    AzureSasCredentials,
    MongoDbCredentials,
    PostgresCredentials,
    S3Credentials,
    SnowflakeCredentials,
    TeradataCredentials,
)
from .utils import Utils


class CredentialsHelper:
    @staticmethod
    def set_credentials(request, raw_data_location, credentials):
        source = getattr(raw_data_location, raw_data_location.WhichOneof("source"))
        if isinstance(source, pb.CSVFileSpec):
            url_path = raw_data_location.csv.path
            CredentialsHelper.get_credentials_based_on_cloud(request, url_path, credentials)
        elif isinstance(source, pb.CSVFolderSpec):
            url_path = raw_data_location.csv_folder.root_folder
            CredentialsHelper.get_credentials_based_on_cloud(request, url_path, credentials)
        elif isinstance(source, pb.JSONFolderSpec):
            url_path = raw_data_location.json_folder.root_folder
            CredentialsHelper.get_credentials_based_on_cloud(request, url_path, credentials)
        elif isinstance(source, pb.JSONFileSpec):
            url_path = raw_data_location.json.path
            CredentialsHelper.get_credentials_based_on_cloud(request, url_path, credentials)
        elif isinstance(source, pb.ParquetFileSpec):
            url_path = raw_data_location.parquet.path
            CredentialsHelper.get_credentials_based_on_cloud(request, url_path, credentials)
        elif isinstance(source, pb.ParquetFolderSpec):
            url_path = raw_data_location.parquet_folder.root_folder
            CredentialsHelper.get_credentials_based_on_cloud(request, url_path, credentials)
        elif isinstance(source, pb.DeltaTableSpec):
            url_path = raw_data_location.delta_table.path
            CredentialsHelper.get_credentials_based_on_cloud(request, url_path, credentials)
        elif isinstance(source, pb.SnowflakeTableSpec):
            CredentialsHelper.set_snowflake_credentials(request, credentials)
        elif isinstance(source, pb.JDBCTableSpec):
            CredentialsHelper.set_jdbc_credentials(request, source.connection_url, credentials)
        elif isinstance(source, pb.TempParquetFileSpec):
            pass
        elif isinstance(source, pb.MongoDbCollectionSpec):
            CredentialsHelper.set_mongodb_credentials(request, credentials)
        else:
            raise Exception("Unsupported external data spec!")

    @staticmethod
    def get_credentials_based_on_cloud(request, url_path, credentials):
        if url_path.lower().startswith("s3"):
            CredentialsHelper.set_s3_credentials(request, credentials)
        elif url_path.lower().startswith("wasb") or url_path.lower().startswith("abfs"):
            CredentialsHelper.set_azure_credentials(request, url_path, credentials)
        else:
            raise Exception("Unsupported external data spec!")

    @staticmethod
    def set_azure_credentials(request, url_path, credentials):
        sas_container = urlparse(url_path).netloc.split("@")[0]
        if credentials is None:
            account_name = Utils.read_env("AZURE_ACCOUNT_NAME", "Azure")
            account_key = os.getenv("AZURE_ACCOUNT_KEY")
            sas_token = os.getenv("AZURE_SAS_TOKEN")
            sp_client_id = os.getenv("AZURE_SP_CLIENT_ID")
            sp_tenant_id = os.getenv("AZURE_SP_TENANT_ID")
            sp_secret = os.getenv("AZURE_SP_SECRET")
            if account_key:
                credentials = AzureKeyCredentials(account_name, account_key)
            elif sas_token:
                credentials = AzureSasCredentials(account_name, sas_token)
            elif sp_client_id and sp_tenant_id and sp_secret:
                credentials = AzurePrincipalCredentials(account_name, sp_client_id, sp_tenant_id, sp_secret)
            else:
                raise Exception(
                    "Either Azure Key, SAS token or Service Credentials environment variable must be specified "
                    "to read from Azure data source!"
                )
        elif not isinstance(
            credentials,
            (AzureKeyCredentials, AzureSasCredentials, AzurePrincipalCredentials),
        ):
            raise Exception(
                "Credentials are not of type AzureKeyCredentials, AzureSasCredentials or AzurePrincipalCredentials!"
            )

        request.cred.azure.account_name = credentials.account_name
        if isinstance(credentials, AzureKeyCredentials):
            request.cred.azure.account_key = credentials.account_key
        if isinstance(credentials, AzureSasCredentials):
            request.cred.azure.sas_token = credentials.sas_token
            request.cred.azure.sas_container = sas_container
        if isinstance(credentials, AzurePrincipalCredentials):
            request.cred.azure.sp_client_id = credentials.client_id
            request.cred.azure.sp_tenant_id = credentials.tenant_id
            request.cred.azure.sp_secret = credentials.secret

    @staticmethod
    def set_s3_credentials(request, credentials):
        if credentials is None:
            if os.environ.get("S3_ACCESS_KEY"):
                access_key = Utils.read_env("S3_ACCESS_KEY", "S3")
            else:
                access_key = None
            if os.environ.get("S3_SECRET_KEY"):
                secret_key = Utils.read_env("S3_SECRET_KEY", "S3")
            else:
                secret_key = None
            if os.environ.get("S3_REGION"):
                region = Utils.read_env("S3_REGION", "S3")
            else:
                region = None
            if os.environ.get("S3_ROLE_ARN"):
                role_arn = Utils.read_env("S3_ROLE_ARN", "S3")
            else:
                role_arn = None
            if os.environ.get("S3_ENDPOINT"):
                endpoint = Utils.read_env("S3_ENDPOINT", "S3")
            else:
                endpoint = None
            if region and (role_arn or (access_key and secret_key)):
                credentials = S3Credentials(access_key, secret_key, region, endpoint, role_arn)
            elif os.path.isfile(os.path.join(os.path.expanduser("~"), ".aws/credentials")):
                config = configparser.RawConfigParser()
                config.read(os.path.join(os.path.expanduser("~"), ".aws/credentials"))
                try:
                    profile = Utils.read_env("AWS_PROFILE", "S3")
                except Exception:
                    logging.warning("AWS_PROFILE is not found in the environment. Profile is set to 'default'.")
                    profile = "default"
                    pass
                aws_key = config.get(profile, "aws_access_key_id")
                aws_secret = config.get(profile, "aws_secret_access_key")
                config.read(os.path.join(os.path.expanduser("~"), ".aws/config"))
                try:
                    aws_region = config.get("default", "region")
                except configparser.NoOptionError:
                    logging.warning("AWS default region is not configured. Using 'us-east-1'.")
                    aws_region = "us-east-1"
                credentials = S3Credentials(access_key=aws_key, secret_key=aws_secret, region=aws_region)
            else:
                request.cred.aws.access_token = "PUBLIC_ONLY"
                request.cred.aws.secret_token = "PUBLIC_ONLY"
                return
        if not isinstance(credentials, S3Credentials):
            raise Exception("Credentials is not of type S3Credentials!")
        request.cred.aws.access_token = credentials.access_token
        request.cred.aws.secret_token = credentials.secret_token
        if credentials.region:
            request.cred.aws.region = credentials.region
        if credentials.role_arn:
            request.cred.aws.role_arn = credentials.role_arn
        if credentials.endpoint:
            request.cred.aws.endpoint = credentials.endpoint

    @staticmethod
    def set_snowflake_credentials(request, credentials):
        if credentials is None:
            credentials = SnowflakeCredentials(
                user=Utils.read_env("SNOWFLAKE_USER", "Snowflake"),
                password=Utils.read_env("SNOWFLAKE_PASSWORD", "Snowflake"),
            )
        elif not isinstance(credentials, SnowflakeCredentials):
            raise Exception("Credentials are not of type snowflakeCredentials!")
        request.cred.snowflake.user = credentials.user
        request.cred.snowflake.password = credentials.password

    @staticmethod
    def set_jdbc_credentials(request, connection_url, credentials):
        database_type = connection_url.split(":")[1]
        if credentials is None:
            if database_type == "teradata":
                credentials = TeradataCredentials(
                    user=Utils.read_env("JDBC_TERADATA_USER", "JDBC Teradata"),
                    password=Utils.read_env("JDBC_TERADATA_PASSWORD", "JDBC Teradata"),
                )
            elif database_type == "postgres":
                credentials = PostgresCredentials(
                    user=Utils.read_env("JDBC_POSTGRES_USER", "JDBC Postgres"),
                    password=Utils.read_env("JDBC_TERADATA_PASSWORD", "JDBC Postgres"),
                )
            else:
                raise Exception("Invalid database type, supported types are: teradata, postgres")
        elif not isinstance(credentials, (TeradataCredentials, PostgresCredentials)):
            raise Exception("Credentials are not of type TeradataCredentials or PostgresCredentials!")
        request.cred.jdbc_database.user = credentials.user
        request.cred.jdbc_database.password = credentials.password

    @staticmethod
    def set_mongodb_credentials(request, credentials):
        if credentials is None:
            credentials = MongoDbCredentials(
                user=Utils.read_env("MONGODB_USER", "MongoDB"),
                password=Utils.read_env("MONGODB_PASSWORD", "MongoDB"),
            )
        elif not isinstance(credentials, MongoDbCredentials):
            raise Exception("Credentials are not of type MongoDbCredentials!")
        request.cred.mongo_db.user = credentials.user
        request.cred.mongo_db.password = credentials.password



import ai.h2o.featurestore.api.v1.CoreService_pb2 as pb

from ..access_type import AccessType
from ..utils import Utils
from .feature_set import FeatureSet


class FeatureSetPopularity:
    def __init__(self, stub, popular_feature_set):
        self._stub = stub
        self._popular_feature_set = popular_feature_set

    @property
    def name(self):
        return self._popular_feature_set.feature_set_name

    @property
    def description(self):
        return self._popular_feature_set.feature_set_description

    @property
    def number_of_retrievals(self):
        return self._popular_feature_set.number_of_retrievals

    @property
    def current_permission(self):
        return AccessType.from_proto_active_permission(self._popular_feature_set.permission)

    def get_feature_set(self):
        request = pb.FeatureSetRequestId(feature_set_id=self._popular_feature_set.feature_set_id)
        response = self._stub.GetFeatureSetById(request)
        return FeatureSet(self._stub, response.feature_set)

    def __repr__(self):
        return Utils.pretty_print_proto(self._popular_feature_set)



import logging
from typing import Iterator

from google.protobuf.empty_pb2 import Empty

import ai.h2o.featurestore.api.v1.CoreService_pb2 as pb

from ..entities.feature_set import FeatureSet
from ..entities.project import Project


class Projects:
    def __init__(self, stub):
        self._stub = stub
        self._default_lock_value = None

    def list(self) -> Iterator[Project]:
        """List all available projects.

        Returns:
            Generator of all projects

        Typical example:
            client.projects.list()

        For more details:
            https://docs.h2o.ai/feature-store/latest-stable/docs/api/projects_api.html#listing-projects
        """
        request = pb.ListProjectsPageRequest()
        while request:
            response = self._stub.ListProjectsPage(request)
            if response.next_page_token:
                request = pb.ListProjectsPageRequest()
                request.page_token = response.next_page_token
            else:
                request = None
            for project in response.projects:
                yield Project(self._stub, project)

    def list_feature_sets(self, project_names=[]) -> Iterator[FeatureSet]:
        """List feature sets across multiple projects.

        Args:
            project_names: (list[str]) A collection of existing project names.

        Returns:
            Iterator[FeatureSet]: An iterator which obtains the feature sets lazily.

        Typical example:
            client.projects.list_feature_sets(["project_name_A", "project_name_B"])

        For more details:
            https://docs.h2o.ai/feature-store/latest-stable/docs/api/projects_api.html#listing-feature-sets-across-multiple-projects
        """
        request = pb.ListFeatureSetsPageRequest()
        request.project_names.extend(project_names)
        while request:
            response = self._stub.ListFeatureSetsPage(request)
            if response.next_page_token:
                request.page_token = response.next_page_token
            else:
                request = None
            for feature_set in response.feature_sets:
                yield FeatureSet(self._stub, feature_set)

    def create(
        self,
        project_name: str,
        description: str = "",
        secret: bool = False,
        locked: bool = None,
    ) -> Project:
        """Create a project.

        Args:
            project_name: (str) A project name.
            description: (str) A description about the project.
            secret: (bool) If True, project is visible only to its owner.
            locked: (bool) If True, only users with consumer permission can list and get feature sets from this project.

        Returns:
            Project: A new project with specified attributes.

        Typical example:
            project = client.projects.create(project_name="project", description="description",
              secret=False, locked=None)

        For more details:
            https://docs.h2o.ai/feature-store/latest-stable/docs/api/projects_api.html#create-a-project
        """
        if locked is None and self._default_lock_value is None:
            self._default_lock_value = self._stub.GetProjectsDefault(Empty()).locked
        if locked is None:
            locked_value = self._default_lock_value
        else:
            locked_value = locked
        request = pb.CreateProjectRequest()
        request.secret = secret
        request.project_name = project_name
        request.description = description
        request.locked = locked_value
        response = self._stub.CreateProject(request)
        if response.already_exists:
            logging.warning("Project '" + project_name + "' already exists.")
        return Project(self._stub, response.project)

    def get(self, project_name: str) -> Project:
        """Obtain an existing project.

        Args:
            project_name: (str) A project name.

        Returns:
            Project: An existing project.

        Typical example:
            project = client.projects.get(project_name="project")
        """
        request = pb.GetProjectRequest()
        request.project_name = project_name
        response = self._stub.GetProject(request)
        return Project(self._stub, response.project)

    def __repr__(self):
        return "This class wraps together methods working with projects"



from typing import Iterator

from google.protobuf.empty_pb2 import Empty

import ai.h2o.featurestore.api.v1.CoreService_pb2 as pb
from featurestore.core.entities.feature_set_popularity import FeatureSetPopularity
from featurestore.core.entities.pinned_feature_set import PinnedFeatureSet
from featurestore.core.entities.recently_used_feature_set import RecentlyUsedFeatureSet
from featurestore.core.entities.recently_used_project import RecentlyUsedProject


class Dashboard:
    def __init__(self, stub):
        self._stub = stub

    def get_feature_sets_popularity(self):
        """Get popular feature sets.

        Returns:
            List of feature sets popularity

        Typical example:
            fs_popularity = client.dashboard.get_feature_sets_popularity()
        """
        response = self._stub.GetFeatureSetsPopularity(Empty())
        return [FeatureSetPopularity(self._stub, popular_feature_set) for popular_feature_set in response.feature_sets]

    def get_recently_used_projects(self):
        """Get projects that were recently utilized.

        Returns:
            List of references to projects

        Typical example:
            recently_used_projects = client.dashboard.get_recently_used_projects()
        """
        response = self._stub.GetRecentlyUsedProjects(Empty())
        return [RecentlyUsedProject(self._stub, project) for project in response.projects]

    def get_recently_used_feature_sets(self):
        """Get feature sets that were recently utilized.

        Returns:
            List of references to feature sets

        Typical example:
            recently_used_feature_sets = client.dashboard.get_recently_used_feature_sets()
        """
        response = self._stub.GetRecentlyUsedFeatureSets(Empty())
        return [RecentlyUsedFeatureSet(self._stub, feature_set) for feature_set in response.feature_sets]

    def list_pinned_feature_sets(self) -> Iterator[PinnedFeatureSet]:
        """List feature sets that were pinned by current user.

        Returns:
            Iterator[PinnedFeatureSet]: An iterator which obtains the pinned feature sets lazily.

        Typical example:
            client.dashboard.list_pinned_feature_sets()
        """
        request = pb.PinnedFeatureSetsRequest()
        while request:
            response = self._stub.ListPinnedFeatureSets(request)
            if response.next_page_token:
                request.page_token = response.next_page_token
            else:
                request = None
            for feature_set in response.feature_sets:
                yield PinnedFeatureSet(self._stub, feature_set)



from typing import Iterator

import ai.h2o.featurestore.api.v1.CoreService_pb2 as pb
from featurestore.core.entities.feature_set_review import FeatureSetReviewRequest, FeatureSetUserReview


class FeatureSetReviews:
    def __init__(self, stub):
        self._stub = stub

    def manageable_requests(self, filters=None) -> Iterator[FeatureSetReviewRequest]:
        """List pending manageable feature set review requests.

        Args:
            (ReviewStatuses) Object represents a specific review status.
            filters: (list[ReviewStatuses]) Filter includes the status of review
            (either IN_PROGRESS, APPROVED or REJECTED).

        Returns:
            Generator of feature set review requests

        Typical example:
            filters = [ReviewStatuses.IN_PROGRESS]
            reviews = client.feature_set_reviews.manageable_requests()

        For more details:
            https://docs.h2o.ai/feature-store/latest-stable/docs/api/feature_set_review_api#manage-review-requests-from-other-users
        """
        request = pb.ListFeatureSetsToReviewRequest(filters=filters)
        return (
            FeatureSetReviewRequest(self._stub, entry)
            for entry in self.__paged_response_to_request(request, self._stub.ListFeatureSetsToReview)
        )

    def my_requests(self, filters=None) -> Iterator[FeatureSetUserReview]:
        """List existing feature set review requests belonging to the user.

        Args:
            (ReviewStatuses) Object represents a specific review status.
            filters: (list[ReviewStatuses]) Filter includes the status of review
            (either IN_PROGRESS, APPROVED or REJECTED).

        Returns:
            Generator of feature set user review requests

        Typical example:
            filters = [ReviewStatuses.IN_PROGRESS]
            reviews = client.feature_set_reviews.my_requests(filters)

        For more details:
            https://docs.h2o.ai/feature-store/latest-stable/docs/api/feature_set_review_api#manage-own-feature-sets-in-review
        """
        request = pb.ListFeatureSetReviewsPageRequest(filters=filters)
        return (
            FeatureSetUserReview(self._stub, entry)
            for entry in self.__paged_response_to_request(request, self._stub.ListFeatureSetsReviewsPage)
        )

    def __paged_response_to_request(self, request, call):
        while request:
            response = call(request)
            if response.next_page_token:
                request.page_token = response.next_page_token
            else:
                request = None
            for entry in response.entries:
                yield entry



from ..utils import Utils
from .base_job import BaseJob


class MaterializationOnlineJob(BaseJob):
    def _response_method(self, job_id):
        response = self._stub.GetMaterializationOnlineJobOutput(job_id)
        return MaterializationOnlineResponse(response)


class MaterializationOnlineResponse:
    def __init__(self, response):
        self._response = response
        self._materialization_scope = response.materialization_scope
        self._num_records = response.num_records

    def get_materialization_scope(self):
        return self._materialization_scope

    def get_num_records(self):
        return self._num_records

    def __repr__(self):
        return Utils.pretty_print_proto(self._response)
